/*
 * light.c - This file contains declarations and defines for light sources.
 *
 *  $Id: light.c,v 1.15 1998/11/30 00:14:36 johns Exp $
 */

#include "machine.h"
#include "types.h"
#include "macros.h"
#include "vector.h"
#include "intersect.h"
#include "util.h"

#define LIGHT_PRIVATE
#include "light.h"

static object_methods light_methods = {
  (void (*)(void *, void *))(light_intersect),
  (void (*)(void *, void *, void *, void *))(light_normal),
  light_bbox, 
  free 
};

point_light * newlight(void * tex, vector ctr, flt rad) {
  point_light * l;
  
  l=(point_light *) malloc(sizeof(point_light));
  memset(l, 0, sizeof(point_light));
  l->methods = &light_methods;

  l->tex=tex;
  l->tex->islight=1;
  l->tex->shadowcast=1;
  l->tex->diffuse=0.0;
  l->tex->specular=0.0;
  l->tex->opacity=1.0;

  l->ctr=ctr;
  l->rad=rad;

  l->attenuationfunc = light_no_attenuation;
  l->Kc = 1.0;
  l->Kl = 0.0;
  l->Kq = 0.0;

  l->spotfunc = light_no_falloff;
  l->spotdir.x = 0.0;
  l->spotdir.y = 0.0;
  l->spotdir.z = 1.0;
  l->fallstart = 0.3;
  l->fallend   = 0.7;

  return l;
}

point_light * newspotlight(void * tex, vector ctr, flt rad, vector dir,
                           flt fallstart, flt fallend) {
  point_light * l;

  l=(point_light *) malloc(sizeof(point_light));
  memset(l, 0, sizeof(point_light));
  l->methods = &light_methods;

  l->tex=tex;
  l->tex->islight=1;
  l->tex->shadowcast=1;
  l->tex->diffuse=0.0;
  l->tex->specular=0.0;
  l->tex->opacity=1.0;

  l->ctr=ctr;
  l->rad=rad;

  l->attenuationfunc = light_no_attenuation;
  l->Kc = 1.0;
  l->Kl = 0.0;
  l->Kq = 0.0;

  l->spotfunc  = light_spotlight_falloff;
  l->spotdir   = dir; 
  l->fallstart = fallstart;
  l->fallend   = fallend;

  return l;
}


void light_set_attenuation(point_light * li, flt Kc, flt Kl, flt Kq) {
  li->Kc = Kc;
  li->Kl = Kl;
  li->Kq = Kq;
  li->attenuationfunc = light_complex_attenuation;
}


/*
 * Private / Internal lighting routines.
 */

static int light_bbox(void * obj, vector * min, vector * max) {
  return 0; /* lights are unbounded currently */
}

static void light_intersect(point_light * l, ray * ry) {
  flt b, disc, t1, t2, temp;
  vector V;

  /* Lights do not cast shadows.. */
  if (ry->flags & RT_RAY_SHADOW)
    return;

  VSUB(l->ctr, ry->o, V);
  VDOT(b, V, ry->d); 
  VDOT(temp, V, V);  

  disc=b*b + l->rad*l->rad - temp;

  if (disc<=0.0) return;
  disc=sqrt(disc);

  t2=b+disc;
  if (t2 <= SPEPSILON) 
    return;
  ry->add_intersection(t2, (object *) l, ry);  

  t1=b-disc;
  if (t1 > SPEPSILON) 
    ry->add_intersection(t1, (object *) l, ry);  
}

static void light_normal(point_light * l, vector * pnt, ray * incident, vector * N) {
  flt len;

  N->x = pnt->x - l->ctr.x;
  N->y = pnt->y - l->ctr.y;
  N->z = pnt->z - l->ctr.z;

  len = sqrt(N->x*N->x + N->y*N->y + N->z*N->z);
  N->x /= len;
  N->y /= len;
  N->z /= len;
}



/*
 * Attenuation functions. 
 */

static flt light_no_attenuation(void * vli, flt Llen) {
  return 1.0;
}

static flt light_linear_attenuation(void * vli, flt Llen) {
  return 1.0 / Llen;
}

static flt light_quadratic_attenuation(void * vli, flt Llen) {
  return 1.0 / (Llen*Llen);
}

static flt light_complex_attenuation(void * vli, flt Llen) {
  point_light * li = (point_light *) vli;
  return 1.0 / (li->Kc + li->Kl*Llen + li->Kq*Llen*Llen);
}



/*
 * Falloff functions for spotlights.
 */

static flt light_no_falloff(void * vli, vector * L) {
  return 1.0;
}


static flt light_spotlight_falloff(void * vli, vector * L) {
  point_light * li = (point_light *) vli;
  flt ang, cang;

  VDOT(cang, li->spotdir, (*L)) 

  ang = acos(-cang); 
  
  if (ang > li->fallstart) {
    if (ang > li->fallend) 
      return 0.0;
    else 
      return 1.0 - ((ang - li->fallstart) / (li->fallend - li->fallstart));
  }

  return 1.0;
}


